/**************************************************************************************************

Copyright (c) 2007   Tsunehiko Kato  and  4D2U Project, National Astronomical Observatory of Japan

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

ȉɒ߂ɏ]A{\tgEFAъ֘Ãt@Ciȉu\tgEFAvj̕擾
邷ׂĂ̐lɑ΂A\tgEFA𖳐ɈƂ𖳏ŋ܂Bɂ́A\tgEFA̕
gpAʁAύXAAfځAЕzATuCZXA/܂͔̔錠Aу\tgEFA
񋟂鑊ɓƂ錠Ɋ܂܂܂B

L̒쌠\і{\A\tgEFÂׂĂ̕܂͏dvȕɋLڂ̂Ƃ
B

\tgEFÁû܂܁vŁAł邩Öقł邩킸A̕ۏ؂Ȃ񋟂܂B
łۏ؂Ƃ́AiA̖ړIւ̓KAьNQɂĂ̕ۏ؂܂݂܂AɌ
̂ł͂܂B ҂܂͒쌠҂́A_sׁAs@sׁA܂͂ȊOł낤ƁA\t
gEFAɋN܂͊֘AA邢̓\tgEFA̎gp܂͂̑̈ɂĐ؂̐A
QȂ̋`ɂĉ̐ӔCȂ̂Ƃ܂B
***************************************************************************************************/
#include  "mitaka.h"
#include  "mtk_galaxies.h"
#include  "mtk_switches.h"
#include  "mtk_font.h"
#include  "mtk_OpenGL.h"
#include  "mtk_gl_common.h"
#include  "mtk_colors.h"
#include  "mtk_file.h"
#include  "mtk_our_galaxy_rt.h"
#include  "mtk_elliptical_galaxy.h"
#include  "../common/astro/astrometry.h"
#include  <vector>
#include  <algorithm>


//-----------
//  Externs
//-----------
extern vector3d  fixed_target_pos[6];
extern float ObjectNameSize;

//
static int SDSS_Thinning = 0;

void
SetDistantGalaxyThinning(int val)
{
	if (val < 0)  val = 0;
	if (val > 10) val = 10;
	SDSS_Thinning = val;
}

//--------------------------------------------
//  ߖT͂̃J^OԍƔzԍ̎Qƃe[u
//--------------------------------------------
static std::map<int, int>	NGC_table, UGC_table, PGC_table, Tully_table, undefined_table;



//*********
enum {
	GALAXY_MTY_UNDETERMINED,
	GALAXY_MTY_ELLIPTICAL,
	GALAXY_MTY_S0,
	GALAXY_MTY_SPIRAL,
	GALAXY_MTY_IRREGULAR,
	GALAXY_MTY_LMC,
	GALAXY_MTY_SMC,
	GALAXY_MTY_M31
};

struct galaxy {
	RA_DEC		eqPos;			// Right ascension/Declination (in degree)
	float		distance;		//  (in pc)
	char		type;			// Morphological type
};


struct nearby_galaxy  :  public galaxy
{
	float		D;				// diameter (in arcmin)
	float		d_D;			// Axial ratio of minor to major diameter
	char		group;			// Group
	char		chCatalogue;	// 'N':NGC, 'I':I, 'U':UCG, ' ':None
	int			Catalogue_no;	//
};


//---------------
//  Galaxy data
//---------------
struct nearby_galaxy2 : public nearby_galaxy
{
	float		R;		// a (in pc)
	vector3f	xyz;	// W

	int	 name_idx;
	bool name_select;
};

struct distant_galaxy
{
	RA_DEC		eqPos;		// Right ascension/Declination (in degree)
	float		distance;	//  (in pc)
	vector3f	xyz;		// W
	float		ecls;		// e-class
};

struct QSO
{
	RA_DEC		eqPos;		// Right ascension/Declination (in degree)
	float		distance;	//  (in pc)
	vector3f	xyz;		// W
	float		z;			// red shift
};


//----------------
//  ߖT̖͂O
//----------------
struct nameBlg {
	string strJ;
	string strE;
};

//----------------------
//  zlist
//  ߖT͂ z \[gp
//----------------------
class zlist
{
public:
	int		ID;		// ID
	float	z;		// z-l

	bool operator<(const zlist& zl) const  { return z < zl.z; }
};

std::vector<zlist>  vZlist(0);



const float galaxy_B_V[5] = { 0.0, 0.96, 0.85, 0.57, 0.27 };
RGBf  galaxy_color[5];
GLuint	texGalaxy;
GLuint	dlGalaxy;

vector<nearby_galaxy2>			NGLX;		// Nearby Galaxies
vector<nameBlg>					NBG_names;	// Names of Nearby Galaxies
vector<vector<distant_galaxy> >	FG_SDSS;	// SDSS Galaxies
vector<vector<QSO> >			vQSO;		// SDSS QSOs


//-------------------
//  Nearby galaxies
//-------------------
int
read_a_nearby_galaxy_record(FILE* fp, nearby_galaxy2& gl)
{
	//--- read data
	int count;
	count = fread(&gl.eqPos.RAdeg,		sizeof(gl.eqPos.RAdeg),			1, fp);
	if (count < 1)  return FILE_FAILED;
	fread(&gl.eqPos.DEdeg,	sizeof(gl.eqPos.DEdeg),		1, fp);
	fread(&gl.distance,		sizeof(gl.distance),		1, fp);
	fread(&gl.D,			sizeof(gl.D),				1, fp);
	fread(&gl.d_D,			sizeof(gl.d_D),				1, fp);
	fread(&gl.type,			sizeof(gl.type),			1, fp);
	fread(&gl.group,		sizeof(gl.group),			1, fp);
	fread(&gl.chCatalogue,	sizeof(gl.chCatalogue),		1, fp);
	fread(&gl.Catalogue_no,	sizeof(gl.Catalogue_no),	1, fp);

	//--- a (in pc)
	gl.R = 0.5 * (pi/180.0 * 1.0/60.0) * gl.D * gl.distance;

	//--- Vʏ̈ʒuiPʔaj
	vector3f Cxyz = gl.eqPos.get_dirc_vec();

	//--- SW
	gl.xyz  = M_R2E * ( gl.distance * Cxyz );

	return  FILE_SUCCESS;
}


bool
read_nearby_galaxy_data(FILE* fp)
{
	int  res;
	nearby_galaxy2 gl;
	int name_idx = 0;

	while (true) {
		res = read_a_nearby_galaxy_record(fp, gl);
		if (res == FILE_FAILED)		break;
		if (res == FILE_SUCCESS)	{
		
			bool bName = true;
			bool bNameSelect = true;
			nameBlg nm;

			//******
			if (gl.type == GALAXY_MTY_M31) {
				//--  Ah_
				fixed_target_pos[3] = gl.xyz;
			} else if (gl.chCatalogue == 'N' && gl.Catalogue_no ==4486) {
				//--  M87
				fixed_target_pos[4] = gl.xyz;
			}
			//******

			gl.name_idx = -1;
			gl.name_select = false;
			NGLX.push_back(gl);

			//---  Qƕ\ɓo^
			int idx = NGLX.size() - 1;
			int cat_no = gl.Catalogue_no;
			switch (gl.chCatalogue) {
				case 'N':	//  NGC
					NGC_table.insert( pair<int, int>(cat_no, idx) );
					break;
				case 'U':	//  UGC
					UGC_table.insert( pair<int, int>(cat_no, idx) );
					break;
				case 'P':	// PGC
					PGC_table.insert( pair<int, int>(cat_no, idx) );
					break;
				case 'T':	// Tully
					Tully_table.insert( pair<int, int>(cat_no, idx) );
					break;
				case '*':	// Undefined
					undefined_table.insert( pair<int, int>(cat_no, idx) );
					break;
				default:
					break;
			}
		}
	}

	vZlist.resize( NGLX.size() );

	return  true;
}



//----------------
//  O̓ǂݍ
//----------------
void
load_nearby_galaxy_names(FILE* fp)
{
	const int BUF_SIZE = 100;
	char buf[BUF_SIZE], buf2[BUF_SIZE];
	char name_E[20], name_J[70];

	while (true) {

		//-- Read one line
		if ( !read_one_line(fp, buf, BUF_SIZE) ) break;

		//-- Catalogue type
		char cat_type = buf[0];

		//-- Catalogue number
		int no;
		if ( !extract_string(buf, 1, 7, buf2) ) continue;
		sscanf(buf2, "%ld", &no);

		//-- Index
		int idx;
		switch (cat_type) {
			case 'N':	// NGC
				{
					map<int, int>::const_iterator it = NGC_table.find( no );
					if (it==NGC_table.end())  continue;
					idx = it->second;
				}
				break;
			case 'U':	// UGC
				{
					map<int, int>::const_iterator it = UGC_table.find( no );
					if (it==UGC_table.end())  continue;
					idx = it->second;
				};
				break;
			case 'P':	// PGC
				{
					map<int, int>::const_iterator it = PGC_table.find( no );
					if (it==PGC_table.end())  continue;
					idx = it->second;
				};
				break;
			case 'T':	// Tully
				{
					map<int, int>::const_iterator it = Tully_table.find( no );
					if (it==Tully_table.end())  continue;
					idx = it->second;
				};
				break;
			case '*':	// Undefined
				{
					map<int, int>::const_iterator it = undefined_table.find( no );
					if (it==undefined_table.end())  continue;
					idx = it->second;
				};
				break;
			default:
				continue;
		}

		//-- Name
		bool res_E, res_J;
		res_E = extract_string(buf, 10, 29, name_E);
		res_J = extract_string(buf, 30, 90, name_J);
		if (!res_E)  continue;
		if (!res_J)  strcpy(name_J, name_E);

		nameBlg nm;
		nm.strE = name_E;
		nm.strJ = name_J;
		NBG_names.push_back(nm);

		NGLX[idx].name_idx = NBG_names.size() - 1;

		//-- Select
		bool bSelect;
		bSelect = (buf[9]!=' ');
		NGLX[idx].name_select = bSelect;
	}
}


bool
load_nearby_galaxies(const directory& dir)
{
	FILE* fp = dir.fopen("NBG.dat", "rb");
	if (fp == NULL)  return false;

	bool res = read_nearby_galaxy_data(fp);
	fclose(fp);
	if ( !res ) return  false;

	fp = dir.fopen("nearby_galaxies_name.dat", "rt");
	if (fp != NULL) {
		load_nearby_galaxy_names(fp);
		fclose(fp);
	}

	return  res;
}


//---  Position  ---
vector3d
get_nearby_galaxy_position(int idx)
{
	return  NGLX[idx].xyz;
}


//--------------------
//  Distant galaxies
//--------------------
bool
load_distant_galaxies(const directory& dir)
{
	const int class_N = 10;
	const float ecls_min = -0.35;
	const float ecls_max = 0.5;
	const float d_ecls = ecls_max - ecls_min;

	FG_SDSS.clear();
	FG_SDSS.resize( class_N );

	FILE* fp = dir.fopen("SDSS_GAL.dat", "rb");
	if (fp == NULL)  return false;

	int count = 0;
	while (true) {
		distant_galaxy gl;
		
		int n = fread(&gl.eqPos.RAdeg, sizeof(gl.eqPos.RAdeg),		1, fp);
		if (n != 1)  break;
		fread(&gl.eqPos.DEdeg,	sizeof(gl.eqPos.DEdeg),		1, fp);
		fread(&gl.distance,		sizeof(gl.distance),		1, fp);
		fread(&gl.ecls,			sizeof(gl.ecls),			1, fp);
		
		//-- ʒu߂
		vector3f Cxyz = gl.eqPos.get_dirc_vec();		// Vʏ̈ʒuiPʔaj
		gl.xyz  = M_R2E * ( gl.distance * Cxyz );		// SW

		//-- NX߂
		int icls = int( (gl.ecls - ecls_min) / d_ecls * class_N);
		if (icls < 0) icls = 0;
		if (icls >= class_N)  icls = class_N - 1;

		FG_SDSS[icls].push_back(gl);
		count++;
	}
	fclose(fp);

	return true;
}


//---  Position  ---
vector3d
get_distant_galaxy_position(int icls, int idx)
{
	return  FG_SDSS[icls][idx].xyz;
}



//--------
//  QSOs
//--------
bool
load_QSOs(const directory& dir)
{
	const int class_N = 20;
	const float z_min = 0.1;
	const float z_max = 1.75;
	const float dz = z_max - z_min;

	vQSO.clear();
	vQSO.resize( class_N );

	FILE* fp = dir.fopen("SDSS_QSO.dat", "rb");
	if (fp == NULL)  return false;

	int count = 0;
	while (true) {
		QSO q;
		
		int n = fread(&q.eqPos.RAdeg, sizeof(q.eqPos.RAdeg),		1, fp);
		if (n != 1)  break;
		fread(&q.eqPos.DEdeg,	sizeof(q.eqPos.DEdeg),	1, fp);
		fread(&q.distance,		sizeof(q.distance),		1, fp);
		fread(&q.z,				sizeof(q.z),			1, fp);
		
		//-- ʒu߂
		vector3f Cxyz = q.eqPos.get_dirc_vec();		// Vʏ̈ʒuiPʔaj
		q.xyz  = M_R2E * ( q.distance * Cxyz );		// SW

		//-- NX߂
		int icls = int( (q.z - z_min) / dz * class_N);
		if (icls < 0) icls = 0;
		if (icls >= class_N)  icls = class_N - 1;

		vQSO[icls].push_back(q);
		count++;
	}
	fclose(fp);

	return true;
}



//------------------
//  ͂̐F̏
//------------------
void
init_galaxy_color()
{
	for (int i=0; i<5; ++i) {
        float Tc = B_V_to_Tc( galaxy_B_V[i] );
		RGBf rgb = blackbody_color(Tc);
		galaxy_color[i] = correct_RGB( rgb );
	}
}



//----------------------------
//  ߖT͂̃eNX`̃[h
//----------------------------
void
load_galaxies_textures()
{
	texGalaxy = LoadTexture("galaxy_class.png", TexShrinkRate);
}


//-------------------------------------------
//  ߖT͂̃C[WpfBXvCXg̏
//-------------------------------------------
void
init_galaxy_display_list()
{
	dlGalaxy = glGenLists(8);
	for (int i=0; i<8; ++i) {
		float du = 1.0f/8.0f;
		float u1 = du*i;
		float u2 = du*(i+1);
		
		glNewList(dlGalaxy+i, GL_COMPILE);
			glBegin(GL_QUADS);
				glTexCoord2f(u2,1);		glVertex3f(1,1,0);
				glTexCoord2f(u1,1);		glVertex3f(-1,1,0);
				glTexCoord2f(u1,0);		glVertex3f(-1,-1,0);
				glTexCoord2f(u2,0);		glVertex3f(1,-1,0);
			glEnd();
		glEndList();
	}
}


//----------------
//  ߖT͂̕`
//----------------
void
draw_nearby_galaxies(const mtk_draw_struct& ds)
{
	const float scale = ds.scale;
	if (tyMode != MODE_HIPPARCOS)  return;

	float af0 = 1.0f;
	if (scale > 1e9) {
		af0 = 1e9/scale;
	} else if (scale < 3e4) {
		af0 = 1 - (3e4 - scale)/(3e4 - 1e4);
	}
	if (af0 < 0.01) {
		if (swGalaxy & GLXY_OUR_GALAXY) {
			draw_our_galaxy_model_rt(ds);
		}
		return;
	}

	if (scale > 2e11)  return;
	const vector3d target_pos = ds.target_pos;
	const vector3d camera_pos = ds.camera.get_pos();
	const frustum  frs = ds.frs;
	const double target_dist_sqr = norm2(target_pos - camera_pos);
	const float Rmax = 0.5*get_Rmax();
	const float Rth  = 0.5*Rmax;

	//-- For culling
	GLdouble MM[16];
	glGetDoublev( GL_MODELVIEW_MATRIX,  MM);

	// make z-list
	int count=0;
    for (int i=0; i<NGLX.size(); ++i) {
		const nearby_galaxy2& gl = NGLX[i];

		//-- Culling
		const vector3f  xyz = gl.xyz - camera_pos;
		if (gl.chCatalogue=='N' && gl.Catalogue_no == 4486) {	// M87
		} else {
		}
		if (xyz.norm() > Rmax)	continue;

		vZlist[count].ID = i;
		vZlist[count].z  = -(MM[14] + MM[2]*xyz.x + MM[6]*xyz.y + MM[10]*xyz.z);	//---->  z = z0 + exx*rc.x + eyz*rc.y + ezz*rc.z;
		++count;
	}

	//---  Our Galaxy  ---
	bool bDrawOurGalaxy = (swGalaxy & GLXY_OUR_GALAXY);
	const vector3f pos_og = M_G2E * vector3f(8000,0,0) - camera_pos;
	const float z_og = -(MM[14] + MM[2]*pos_og.x + MM[6]*pos_og.y + MM[10]*pos_og.z);
	if (count == 0) {
		if (bDrawOurGalaxy) {
			draw_our_galaxy_model_rt(ds);
		}
		return;
	}

	// sort
	sort(vZlist.begin(), vZlist.begin() + count);

	//-- draw
	glEnable(GL_BLEND);
	glDepthMask(GL_FALSE);
	glBindTexture(GL_TEXTURE_2D, texGalaxy);

	float ptsz_scl = 1/(ds.rad_per_pixel / 0.001);

	for (int i=count-1; i>=0; --i) {
		// Our Galaxy
		if (bDrawOurGalaxy && vZlist[i].z < z_og) {
			draw_our_galaxy_model_rt(ds);
			glDisable(GL_DEPTH_TEST);
			glDepthMask(GL_FALSE);
			bDrawOurGalaxy = false;
		}

		const nearby_galaxy2& gl = NGLX[vZlist[i].ID];
		float R_gal = gl.R;		// ͂̔a
		float sz = R_gal * get_scale_factor();
		sz /= 0.8;	// 摜̃TCY̕␳

		const vector3f  xyz = gl.xyz - camera_pos;
		float af = af0;
		const float dist = xyz.norm();
		if (dist > Rth) {
			af = 1 - (dist-Rth)/(Rmax-Rth);
		}
		if (af < 0.01) continue;

		glPushMatrix();
			// ͒Sֈړ
			glTranslatef(xyz.x,xyz.y,xyz.z);

			float M[16];
			glGetFloatv( GL_MODELVIEW_MATRIX, M);
			float r = sqrt( M[12]*M[12] + M[13]*M[13] + M[14]*M[14] );
			float sz2 = R_gal/scale /r*screen_z * 2000;
			if (sz2 < 0.5) sz=2;

			if (sz2 >= 0.5 && gl.chCatalogue=='N' && gl.Catalogue_no == 4486) {
				//  cD 
				draw_elliptical_galaxy(ds);
			} else {
				// ͂`
				if (sz2 >= 0.5) {
					glBlendFunc(GL_ONE, GL_ONE);

					glEnable(GL_TEXTURE_2D);
					if (bVirgoCluster && gl.group==11)	{ glColor4f(0.2,1,0.4,af); }
					else								{ glColor4f(af,af,af,af); }

					toward_eye_pos();
					glScalef(sz,sz,1);
					glCallList(dlGalaxy + gl.type);
					glDisable(GL_TEXTURE_2D);
				} else {
					glDisable(GL_TEXTURE_2D);
					glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
					if (bVirgoCluster && gl.group == 11)	{ glColor4f(0.2,1.0,0.4,af); glPointSize(3.5*ptsz_scl); }	// For Virgo cluster
					else									{ glColor4f(0.9,0.9,0.9,af); glPointSize(3.0*ptsz_scl); }
					
					glBegin(GL_POINTS);
						glVertex3f(0,0,0);
					glEnd();
				}
			}

			// O`
			if ( test_disp_switch(DISPSW_PLANET, DISPM_NAME) && gl.name_idx >= 0) {
				if (gl.name_select || !(swStar & STAR_NAME_SELECT)) {
					double dist_sqr = norm2(gl.xyz - target_pos);
					if (dist_sqr > 1) {
						float aa;
						aa = 200*(dist_sqr / target_dist_sqr);
						float aaa= 0.5*aa - 0.3;
						if (aaa < 1) {
							af *= aaa;
						}
					}

					if (af > 0.01) {
						const RGBf& col = get_color(MTK_COLOR_NAME_OBJECT);
						glColor4f(col.r, col.g, col.b, af);
						glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
					
						float R = 0.5*sz;
						prepare_3Dtext_1(R, 0, ObjectNameSize*ds.ltf, 1.0, false);
						
						if (gl.name_idx >= 0) {
							nameBlg& nm = NBG_names[gl.name_idx];
							if (ds.lang == LANG_ENG) {
								draw_text(nm.strE.c_str());
							} else {
								draw_text(nm.strJ.c_str());
							};
						}
						glBlendFunc(GL_ONE, GL_ONE);
					}
				}
			}
		glPopMatrix();
	}
	glPointSize(1);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_TEXTURE_2D);
	glDepthMask(GL_TRUE);

	if (bDrawOurGalaxy) {
		draw_our_galaxy_model_rt(ds);
	}
}


void
draw_distant_galaxies(const mtk_draw_struct& ds, vector<vector<distant_galaxy> >& FGLX, float pt_sz)
{
	const float scale = ds.scale;
	if (tyMode!=MODE_HIPPARCOS || scale < 1000)  return;
	const vector3d camera_pos = ds.camera.get_pos();

	//-- For culling
	GLdouble MM[16];
	glGetDoublev( GL_MODELVIEW_MATRIX,  MM);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_DEPTH_TEST);

	int skp = 1;
	float af = 1.0;
	pt_sz /= ds.rad_per_pixel / 0.001;
	if (pt_sz < 1) {
		af = pt_sz;
		pt_sz = 1;
	}
	glPointSize(pt_sz);

	// ߋ̃tF[h
	const float scl1 = 5.0e7;
	const float scl2 = 5.0e8;
	const float scl3 = 3.0e9;

	if (scale < scl1)  return;
	if (scale < scl2) {
		const float log_scl = log(scale);	
		const float log_scl1 = log(scl1);
		const float log_scl2 = log(scl2);
		af *= (log_scl - log_scl1) / (log_scl2-log_scl1);
	}

	// ̃tF[h
	if (scale > scl3) {
		float L = scale / scl3;

		af /= L*L;
	}
	af *= 0.7;
	if (af <= 0.00001)  return;
	while (af < 0.02) {
		af *= 2;
		skp *= 2;
	}

	// Thinning out
	for (int i=0; i<SDSS_Thinning; i++) {
		af *= 2;
		skp *= 2;
	}

	if (af > 1) {
		af = 1;
	}

	glPushMatrix();
	glTranslated(-camera_pos.x, -camera_pos.y, -camera_pos.z);

	const int class_N = FGLX.size();
	for (int ic=0; ic<class_N; ic++) {
		float t = ic/float(class_N-1);
		float B_V = 2.2*(1-t) - 0.2;
		RGBf col = correct_RGB( 0.9f * blackbody_color( B_V_to_Tc( B_V ) ) );
		glColor4f( col.r, col.g, col.b, af );
		const vector<distant_galaxy>&  GAL = FGLX[ic];
		
		glBegin(GL_POINTS);
		for (int i=0; i<GAL.size(); i+=skp) {
			const distant_galaxy& gl = GAL[i];
			glVertex3fv(gl.xyz.get_pointer());
		}
		glEnd();
	}
	glPopMatrix();


	glEnable(GL_DEPTH_TEST);
	glPointSize(1);
}


void
draw_QSOs(const mtk_draw_struct& ds, float pt_sz)
{
	const float scale = ds.scale;
	if (tyMode!=MODE_HIPPARCOS || scale < 1000)  return;
	const vector3d camera_pos = ds.camera.get_pos();

	//-- For culling
	GLdouble MM[16];
	glGetDoublev( GL_MODELVIEW_MATRIX,  MM);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_DEPTH_TEST);

	int skp = 1;
	float af = 1.0;
	pt_sz /= ds.rad_per_pixel / 0.001;
	if (pt_sz < 1) {
		af = pt_sz;
		pt_sz = 1;
	}
	glPointSize(pt_sz);

	// ߋ̃tF[h
	const float scl1 = 3.0e8;
	const float scl2 = 1.9e10;
	if (scale < scl1)  return;
	if (scale < scl2) {
		const float log_scl = log(scale);	
		const float log_scl1 = log(scl1);
		const float log_scl2 = log(scl2);
		af *= (log_scl - log_scl1) / (log_scl2-log_scl1);
	}

	// ̃tF[h
	if (scale > scl2) {
		float L = scale / scl2;
		af /= L*L;
	}
	af *= 0.7;
	if (af <= 0.00001)  return;
	while (af < 0.02) {
		af *= 2;
		skp *= 2;
	}

	// Thinning out
	for (int i=0; i<SDSS_Thinning; i++) {
		af *= 2;
		skp *= 2;
	}

	if (af > 1) {
		af = 1;
	}

	glPushMatrix();
	glTranslated(-camera_pos.x, -camera_pos.y, -camera_pos.z);
	const RGBf col1 = get_color( MTK_COLOR_QSO );
	const RGBf col2 = get_color( MTK_COLOR_QSO );
	const int class_N = vQSO.size();
	const float z_min = 0.1;
	const float z_max = 1.75;
	for (int ic=0; ic<class_N; ic++) {
		float t = ic/float(class_N-1);
		RGBf col = (1-t)*col1 + t*col2;
		glColor4f( col.r, col.g, col.b, af );
		const vector<QSO>&  Q = vQSO[ic];
		
		glBegin(GL_POINTS);
		for (int i=0; i<Q.size(); i+=skp) {
			const QSO& q = Q[i];
			glVertex3fv(q.xyz.get_pointer());
		}
		glEnd();
	}
	glPopMatrix();
	glEnable(GL_DEPTH_TEST);
	glPointSize(1);
}

void
draw_galaxies(const mtk_draw_struct& ds)
{
	if (swGalaxy & GLXY_NEARBY_GALAXY)	{
		draw_nearby_galaxies(ds);	// ͌n͂̒ŕ`悷
	} else {
		if (swGalaxy & GLXY_OUR_GALAXY)	{
			draw_our_galaxy_model_rt(ds);
		}
	}
	if (swGalaxy & GLXY_DISTANT_GALAXY)		draw_distant_galaxies(ds, FG_SDSS, 1.7);
}

void
draw_QSOs(const mtk_draw_struct& ds)
{
	draw_QSOs(ds, 1.7);
}


//----------
//  f[^
//----------
int
get_number_of_nearby_galaxies()
{
	return  NGLX.size();
}

int
get_number_of_distant_galaxies()
{
	int count = 0;
	for (int i=0; i<FG_SDSS.size(); i++) {
		count += FG_SDSS[i].size();
	}

	return  count;
}

int
get_number_of_QSOs()
{
	int count = 0;
	for (int i=0; i<vQSO.size(); i++) {
		count += vQSO[i].size();
	}

	return  count;
}
